from concurrent.futures import ThreadPoolExecutor
from threading import Barrier

import pytest

from zict import InsertionSortedSet
from zict.tests import utils_test


def test_insertion_sorted_set():
    s = InsertionSortedSet()

    assert not s
    assert len(s) == 0
    assert list(s) == []
    assert s == set()
    assert s != []
    assert s == InsertionSortedSet()
    assert 1 not in s
    s.discard(1)
    with pytest.raises(KeyError):
        s.remove(1)
    with pytest.raises(KeyError):
        s.pop()
    with pytest.raises(KeyError):
        s.popleft()
    with pytest.raises(KeyError):
        s.popright()

    s.add(1)
    assert 1 in s
    assert 2 not in s
    assert len(s) == 1
    assert list(s) == [1]
    assert s == {1}
    assert s != [1]
    assert s & {1, 2} == {1}
    assert s | {1, 2} == {1, 2}
    assert s - {1, 2} == set()

    # Add already-existing element
    s.add(1)
    assert len(s) == 1
    assert list(s) == [1]

    s.remove(1)
    assert not s
    s.add(1)
    assert list(s) == [1]
    s.discard(1)
    assert not s
    s.add(1)
    assert s.pop() == 1
    s.add(1)
    s.clear()
    assert not s

    # Initialise from iterable
    s = InsertionSortedSet(iter([3, 1, 2, 5, 4, 6, 0]))
    assert list(s) == [3, 1, 2, 5, 4, 6, 0]

    # Adding already-existing element does not change order
    s.add(2)
    assert list(s) == [3, 1, 2, 5, 4, 6, 0]

    # Removing element does not change order
    s.remove(2)
    assert list(s) == [3, 1, 5, 4, 6, 0]

    s.add(2)  # Re-added elements are added to the end
    s.add(7)
    assert list(s) == [3, 1, 5, 4, 6, 0, 2, 7]

    assert [s.popleft() for _ in range(len(s))] == [3, 1, 5, 4, 6, 0, 2, 7]

    s |= [3, 1, 5, 4, 6, 0, 2, 7]
    assert [s.popright() for _ in range(len(s))] == [7, 2, 0, 6, 4, 5, 1, 3]

    # pop() is an alias to popright()
    s |= [3, 1, 5, 4, 6, 0, 2, 7]
    assert [s.pop() for _ in range(len(s))] == [7, 2, 0, 6, 4, 5, 1, 3]


@pytest.mark.stress
@pytest.mark.repeat(utils_test.REPEAT_STRESS_TESTS)
@pytest.mark.parametrize("method,size", [("popleft", 100_000), ("popright", 5_000_000)])
def test_insertion_sorted_set_threadsafe(method, size):
    s = InsertionSortedSet(range(size))
    m = getattr(s, method)
    barrier = Barrier(2)

    def t():
        barrier.wait()
        n = 0
        prev = -1 if method == "popleft" else size
        while True:
            try:
                v = m()
                assert v > prev if method == "popleft" else v < prev, (v, prev, len(s))
                prev = v
                n += 1
            except KeyError:
                assert not s
                return n

    with ThreadPoolExecutor(2) as ex:
        f1 = ex.submit(t)
        f2 = ex.submit(t)
        # On Linux, these are in the 38_000 ~ 62_000 range.
        # On Windows, we've seen as little as 2300.
        assert f1.result() > 100
        assert f2.result() > 100



from __future__ import annotations

import gc
import sys
import threading
from concurrent.futures import ThreadPoolExecutor

import pytest

try:
    import psutil
except ImportError:
    psutil = None  # type: ignore


@pytest.fixture
def check_fd_leaks():
    if sys.platform == "win32" or psutil is None:
        yield
    else:
        proc = psutil.Process()
        before = proc.num_fds()
        yield
        gc.collect()
        assert proc.num_fds() == before


@pytest.fixture
def is_locked():
    """Callable that returns True if the parameter zict mapping has its RLock engaged"""
    with ThreadPoolExecutor(1) as ex:

        def __is_locked(d):
            out = d.lock.acquire(blocking=False)
            if out:
                d.lock.release()
            return not out

        def _is_locked(d):
            return ex.submit(__is_locked, d).result()

        yield _is_locked


@pytest.fixture
def check_thread_leaks():
    active_threads_start = threading.enumerate()

    yield

    bad_threads = [
        thread for thread in threading.enumerate() if thread not in active_threads_start
    ]
    if bad_threads:
        raise RuntimeError(f"Leaked thread(s): {bad_threads}")



import os
import pathlib

import pytest

from zict import LMDB
from zict.tests import utils_test

pytest.importorskip("lmdb")


@pytest.mark.parametrize("dirtype", [str, pathlib.Path, lambda x: x])
def test_dirtypes(tmp_path, check_fd_leaks, dirtype):
    z = LMDB(tmp_path)
    z["x"] = b"123"
    assert z["x"] == b"123"
    del z["x"]


def test_mapping(tmp_path, check_fd_leaks):
    """
    Test mapping interface for LMDB().
    """
    z = LMDB(tmp_path)
    utils_test.check_mapping(z)


def test_bad_types(tmp_path, check_fd_leaks):
    z = LMDB(tmp_path)
    utils_test.check_bad_key_types(z)
    utils_test.check_bad_value_types(z)


def test_reuse(tmp_path, check_fd_leaks):
    """
    Test persistence of a LMDB() mapping.
    """
    with LMDB(tmp_path) as z:
        assert len(z) == 0
        z["abc"] = b"123"

    with LMDB(tmp_path) as z:
        assert len(z) == 1
        assert z["abc"] == b"123"


def test_creates_dir(tmp_path, check_fd_leaks):
    with LMDB(tmp_path, check_fd_leaks):
        assert os.path.isdir(tmp_path)


def test_file_descriptors_dont_leak(tmp_path, check_fd_leaks):
    z = LMDB(tmp_path)
    del z

    z = LMDB(tmp_path)
    z.close()

    with LMDB(tmp_path) as z:
        pass


def test_map_size(tmp_path, check_fd_leaks):
    import lmdb

    z = LMDB(tmp_path, map_size=2**20)
    z["x"] = b"x" * 2**19
    with pytest.raises(lmdb.MapFullError):
        z["y"] = b"x" * 2**20



from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable, Iterable, Iterator, Mapping, MutableMapping
from typing import Generic, TypeVar

from zict.common import KT, VT, ZictBase, close, discard, flush, locked

MKT = TypeVar("MKT")


class Sieve(ZictBase[KT, VT], Generic[KT, VT, MKT]):
    """Store values in different mappings based on a selector's
    output.

    This creates a MutableMapping combining several underlying
    MutableMappings for storage.  Items are dispatched based on
    a selector function provided by the user.

    Parameters
    ----------
    mappings: dict of {mapping key: MutableMapping}
    selector: callable (key, value) -> mapping key

    Notes
    -----
    If you call methods of this class from multiple threads, access will be fast as long
    as the ``__contains__`` and ``__delitem__`` methods of all underlying mappins are
    fast. ``__getitem__`` and ``__setitem__`` methods of the underlying mappings are not
    protected by locks.

    Examples
    --------
    >>> small = {}
    >>> large = DataBase()                        # doctest: +SKIP
    >>> mappings = {True: small, False: large}    # doctest: +SKIP
    >>> def is_small(key, value):                 # doctest: +SKIP
    ...     return sys.getsizeof(value) < 10000   # doctest: +SKIP
    >>> d = Sieve(mappings, is_small)             # doctest: +SKIP
    """

    mappings: Mapping[MKT, MutableMapping[KT, VT]]
    selector: Callable[[KT, VT], MKT]
    key_to_mapping: dict[KT, MutableMapping[KT, VT]]
    gen: int

    def __init__(
        self,
        mappings: Mapping[MKT, MutableMapping[KT, VT]],
        selector: Callable[[KT, VT], MKT],
    ):
        super().__init__()
        self.mappings = mappings
        self.selector = selector
        self.key_to_mapping = {}
        self.gen = 0

    def __getitem__(self, key: KT) -> VT:
        # Note that this may raise KeyError if you call it in the middle of __setitem__
        # or update for an already existing key
        return self.key_to_mapping[key][key]

    @locked
    def __setitem__(self, key: KT, value: VT) -> None:
        discard(self, key)
        mkey = self.selector(key, value)
        mapping = self.mappings[mkey]
        self.key_to_mapping[key] = mapping
        self.gen += 1
        gen = self.gen

        with self.unlock():
            mapping[key] = value

        if gen != self.gen and self.key_to_mapping.get(key) is not mapping:
            # Multithreaded race condition
            discard(mapping, key)

    @locked
    def __delitem__(self, key: KT) -> None:
        mapping = self.key_to_mapping.pop(key)
        self.gen += 1
        discard(mapping, key)

    @locked
    def _do_update(self, items: Iterable[tuple[KT, VT]]) -> None:
        # Optimized update() implementation issuing a single update()
        # call per underlying mapping.
        updates = defaultdict(list)
        self.gen += 1
        gen = self.gen

        for key, value in items:
            old_mapping = self.key_to_mapping.pop(key, None)
            if old_mapping is not None:
                discard(old_mapping, key)
            mkey = self.selector(key, value)
            mapping = self.mappings[mkey]
            updates[mkey].append((key, value))
            self.key_to_mapping[key] = mapping

        with self.unlock():
            for mkey, mitems in updates.items():
                mapping = self.mappings[mkey]
                mapping.update(mitems)

        if gen != self.gen:
            # Multithreaded race condition
            for mkey, mitems in updates.items():
                mapping = self.mappings[mkey]
                for key, _ in mitems:
                    if self.key_to_mapping.get(key) is not mapping:
                        discard(mapping, key)

    def __len__(self) -> int:
        return len(self.key_to_mapping)

    def __iter__(self) -> Iterator[KT]:
        return iter(self.key_to_mapping)

    def __contains__(self, key: object) -> bool:
        return key in self.key_to_mapping

    def __str__(self) -> str:
        return f"Sieve<{self.mappings}>"

    __repr__ = __str__

    def flush(self) -> None:
        flush(*self.mappings.values())

    def close(self) -> None:
        close(*self.mappings.values())



import asyncio
import contextvars
import threading
import time
from collections import UserDict
from concurrent.futures import Executor, Future

import pytest

from zict import AsyncBuffer, Func
from zict.tests import utils_test


@pytest.mark.asyncio
async def test_simple(check_thread_leaks):
    with AsyncBuffer({}, utils_test.SlowDict(0.01), n=3) as buff:
        buff["a"] = 1
        buff["b"] = 2
        buff["c"] = 3
        assert set(buff.fast) == {"a", "b", "c"}
        assert not buff.slow
        assert not buff.futures

        buff["d"] = 4
        assert set(buff.fast) == {"a", "b", "c", "d"}
        assert not buff.slow
        assert buff.futures
        await asyncio.wait(buff.futures)
        assert set(buff.fast) == {"b", "c", "d"}
        assert set(buff.slow) == {"a"}

        buff.async_evict_until_below_target()
        assert not buff.futures
        buff.async_evict_until_below_target(10)
        assert not buff.futures
        buff.async_evict_until_below_target(2)
        assert buff.futures
        await asyncio.wait(buff.futures)
        assert set(buff.fast) == {"c", "d"}
        assert set(buff.slow) == {"a", "b"}

        # Do not incur in threading sync cost if everything is in fast
        assert list(buff.fast.order) == ["c", "d"]
        future = buff.async_get(["c"])
        assert future.done()
        assert await future == {"c": 3}
        assert list(buff.fast.order) == ["d", "c"]

        # Do not disturb LRU order in case of missing keys
        with pytest.raises(KeyError, match="m"):
            _ = buff.async_get(["d", "m"], missing="raise")
        assert list(buff.fast.order) == ["d", "c"]

        future = buff.async_get(["d", "m"], missing="omit")
        assert future.done()
        assert await future == {"d": 4}
        assert list(buff.fast.order) == ["c", "d"]

        with pytest.raises(ValueError):
            _ = buff.async_get(["a"], missing="misspell")

        # Asynchronously retrieve from slow
        future = buff.async_get(["a", "b"])
        assert not future.done()
        assert future in buff.futures
        assert await future == {"a": 1, "b": 2}
        assert not buff.futures
        assert set(buff.fast) == {"d", "a", "b"}
        assert set(buff.slow) == {"c"}


@pytest.mark.asyncio
async def test_double_evict(check_thread_leaks):
    """User calls async_evict_until_below_target() while the same is already running"""
    with AsyncBuffer({}, utils_test.SlowDict(0.01), n=3) as buff:
        buff["x"] = 1
        buff["y"] = 2
        buff["z"] = 3
        assert len(buff.fast) == 3
        assert not buff.futures

        buff.async_evict_until_below_target(2)
        assert len(buff.futures) == 1
        assert list(buff.evicting.values()) == [2]

        # Evicting to the same n is a no-op
        buff.async_evict_until_below_target(2)
        assert len(buff.futures) == 1
        assert list(buff.evicting.values()) == [2]

        # Evicting to a lower n while a previous eviction is still running does not
        # cancel the previous eviction
        buff.async_evict_until_below_target(1)
        assert len(buff.futures) == 2
        assert list(buff.evicting.values()) == [2, 1]
        await asyncio.wait(buff.futures, return_when=asyncio.FIRST_COMPLETED)
        assert len(buff.futures) == 1
        assert list(buff.evicting.values()) == [1]
        await asyncio.wait(buff.futures)
        assert not buff.futures
        assert not buff.evicting

        assert buff.fast == {"z": 3}
        assert buff.slow.data == {"x": 1, "y": 2}

        # Evicting to negative n while fast is empty does nothing
        buff.evict_until_below_target(0)
        buff.async_evict_until_below_target(-1)
        assert not buff.futures
        assert not buff.evicting


@pytest.mark.asyncio
async def test_close_during_evict(check_thread_leaks):
    buff = AsyncBuffer({}, utils_test.SlowDict(0.01), n=100)
    buff.update({i: i for i in range(100)})
    assert not buff.futures
    assert len(buff.fast) == 100

    buff.async_evict_until_below_target(0)
    while not buff.slow:
        await asyncio.sleep(0.01)
    assert buff.fast
    assert buff.futures

    buff.close()
    await asyncio.wait(buff.futures)
    assert not buff.futures
    assert buff.fast
    assert buff.slow


@pytest.mark.asyncio
async def test_close_during_get(check_thread_leaks):
    buff = AsyncBuffer({}, utils_test.SlowDict(0.01), n=100)
    buff.slow.data.update({i: i for i in range(100)})
    assert len(buff) == 100
    assert not buff.fast

    future = buff.async_get(list(range(100)))
    assert buff.futures
    while not buff.fast:
        await asyncio.sleep(0.01)

    buff.close()
    with pytest.raises(asyncio.CancelledError):
        await future
    await asyncio.wait(buff.futures)
    assert not buff.futures

    assert buff.fast
    assert buff.slow


@pytest.mark.asyncio
async def test_contextvars(check_thread_leaks):
    ctx = contextvars.ContextVar("v", default=0)
    in_dump = threading.Event()
    in_load = threading.Event()
    block_dump = threading.Event()
    block_load = threading.Event()

    def dump(v):
        in_dump.set()
        assert block_dump.wait(timeout=5)
        return v + ctx.get()

    def load(v):
        in_load.set()
        assert block_load.wait(timeout=5)
        return v + ctx.get()

    with AsyncBuffer({}, Func(dump, load, {}), n=0.1) as buff:
        ctx.set(20)  # Picked up by dump
        buff["x"] = 1
        assert buff.futures
        assert in_dump.wait(timeout=5)
        ctx.set(300)  # Changed while dump runs. Won't be picked up until load.
        block_dump.set()
        await asyncio.wait(buff.futures)
        assert buff.slow.d == {"x": 21}
        fut = buff.async_get(["x"])
        assert in_load.wait(timeout=5)
        ctx.set(4000)  # Changed while load runs. Won't be picked up.
        block_load.set()
        assert await fut == {"x": 321}  # 1 + 20 (added by dump) + 300 (added by load)


@pytest.mark.asyncio
@pytest.mark.parametrize("missing", ["raise", "omit"])
async def test_race_condition_get_async_delitem(check_thread_leaks, missing):
    """All required keys exist in slow when you call get_async(); however some are
    deleted by the time the offloaded thread retrieves their values.
    """

    class Slow(UserDict):
        def __getitem__(self, key):
            if key in self:
                time.sleep(0.01)
            return super().__getitem__(key)

    with AsyncBuffer({}, Slow(), n=100) as buff:
        buff.slow.update({i: i for i in range(100)})
        assert len(buff) == 100

        future = buff.async_get(list(range(100)), missing=missing)
        while not buff.fast:
            await asyncio.sleep(0.01)
        assert buff.slow
        # Don't use clear(); it uses __iter__ which would not return until restore is
        # completed
        for i in range(100):
            del buff[i]
        assert not buff.fast
        assert not buff.slow
        assert not future.done()

        if missing == "raise":
            with pytest.raises(KeyError):
                await future
        else:
            out = await future
            assert 0 < len(out) < 100


@pytest.mark.asyncio
async def test_multiple_offload_threads():
    barrier = threading.Barrier(2)

    class Slow(UserDict):
        def __getitem__(self, key):
            barrier.wait(timeout=5)
            return super().__getitem__(key)

    with AsyncBuffer({}, Slow(), n=100, nthreads=2) as buff:
        buff["x"] = 1
        buff["y"] = 2
        buff.evict_until_below_target(0)
        assert not buff.fast
        assert set(buff.slow) == {"x", "y"}

        out = await asyncio.gather(buff.async_get(["x"]), buff.async_get(["y"]))
        assert out == [{"x": 1}, {"y": 2}]


@pytest.mark.asyncio
async def test_external_executor():
    n_submit = 0

    class MyExecutor(Executor):
        def submit(self, fn, /, *args, **kwargs):
            nonlocal n_submit
            n_submit += 1
            out = fn(*args, **kwargs)
            f = Future()
            f.set_result(out)
            return f

        def shutdown(self, *args, **kwargs):
            raise AssertionError("AsyncBuffer.close() called executor.shutdown()")

    ex = MyExecutor()
    buff = AsyncBuffer({}, {}, n=1, executor=ex)
    buff["x"] = 1
    buff["y"] = 2  # Evict x
    assert buff.fast.d == {"y": 2}
    assert buff.slow == {"x": 1}
    assert await buff.async_get(["x"]) == {"x": 1}  # Restore x, evict y
    assert buff.fast.d == {"x": 1}
    assert buff.slow == {"y": 2}
    assert n_submit == 2
    buff.close()



import zipfile
from collections.abc import MutableMapping

import pytest

from zict import Zip
from zict.tests import utils_test


@pytest.fixture
def fn(tmp_path, check_fd_leaks):
    yield tmp_path / "tmp.zip"


def test_simple(fn):
    z = Zip(fn)
    assert isinstance(z, MutableMapping)
    assert not z

    assert list(z) == list(z.keys()) == []
    assert list(z.values()) == []
    assert list(z.items()) == []

    z["x"] = b"123"
    assert list(z) == list(z.keys()) == ["x"]
    assert list(z.values()) == [b"123"]
    assert list(z.items()) == [("x", b"123")]
    assert z["x"] == b"123"

    z.flush()
    zz = zipfile.ZipFile(fn, mode="r")
    assert zz.read("x") == b"123"

    z["y"] = b"456"
    assert z["y"] == b"456"


def test_setitem_typeerror(fn):
    z = Zip(fn)
    with pytest.raises(TypeError):
        z["x"] = 123


def test_contextmanager(fn):
    with Zip(fn) as z:
        z["x"] = b"123"

    zz = zipfile.ZipFile(fn, mode="r")
    assert zz.read("x") == b"123"


def test_missing_key(fn):
    z = Zip(fn)

    with pytest.raises(KeyError):
        z["x"]


def test_close(fn):
    z = Zip(fn)

    z["x"] = b"123"
    z.close()

    zz = zipfile.ZipFile(fn, mode="r")
    assert zz.read("x") == b"123"

    with pytest.raises(IOError):
        z["y"] = b"123"


def test_bytearray(fn):
    data = bytearray(b"123")
    with Zip(fn) as z:
        z["x"] = data

    with Zip(fn) as z:
        assert z["x"] == b"123"


def test_memoryview(fn):
    data = memoryview(b"123")
    with Zip(fn) as z:
        z["x"] = data

    with Zip(fn) as z:
        assert z["x"] == b"123"


def check_mapping(z):
    """Shorter version of utils_test.check_mapping, as zip supports neither update nor
    delete
    """
    assert isinstance(z, MutableMapping)
    utils_test.check_empty_mapping(z)

    z["abc"] = b"456"
    z["xyz"] = b"12"
    assert len(z) == 2
    assert z["abc"] == b"456"

    utils_test.check_items(z, [("abc", b"456"), ("xyz", b"12")])

    assert "abc" in z
    assert "xyz" in z
    assert "def" not in z

    with pytest.raises(KeyError):
        z["def"]


def test_mapping(fn):
    """
    Test mapping interface for Zip().
    """
    with Zip(fn) as z:
        check_mapping(z)
        utils_test.check_closing(z)


def test_no_delete_update(fn):
    with Zip(fn) as z:
        z["x"] = b"123"
        with pytest.raises(NotImplementedError):
            del z["x"]
        with pytest.raises(NotImplementedError):
            z["x"] = b"456"
        assert len(z) == 1
        assert z["x"] == b"123"


def test_bad_types(fn):
    with Zip(fn) as z:
        utils_test.check_bad_key_types(z, has_del=False)
        utils_test.check_bad_value_types(z)



from __future__ import annotations

import random
import string
import threading
import time
from collections import UserDict
from collections.abc import ItemsView, KeysView, MutableMapping, ValuesView
from concurrent.futures import ThreadPoolExecutor

import pytest

from zict.common import ZictBase

# How many times to repeat non-deterministic stress tests.
# You may set it as high as 50 if you wish to run in CI.
REPEAT_STRESS_TESTS = 1


def generate_random_strings(n, min_len, max_len):
    r = random.Random(42)
    out = []
    chars = string.ascii_lowercase + string.digits

    for _ in range(n):
        nchars = r.randint(min_len, max_len)
        s = "".join(r.choice(chars) for _ in range(nchars))
        out.append(s)

    return out


def to_bytestring(s):
    if isinstance(s, bytes):
        return s
    else:
        return s.encode("latin1")


def check_items(z: MutableMapping, expected_items: list[tuple[str, bytes]]) -> None:
    items = list(z.items())
    assert len(items) == len(expected_items)
    assert sorted(items) == sorted(expected_items)
    # All iterators should walk the mapping in the same order
    assert list(z.keys()) == [k for k, v in items]
    assert list(z.values()) == [v for k, v in items]
    assert list(z) == [k for k, v in items]

    # ItemsView, KeysView, ValuesView.__contains__()
    assert isinstance(z.keys(), KeysView)
    assert isinstance(z.values(), ValuesView)
    assert isinstance(z.items(), ItemsView)
    assert items[0] in z.items()
    assert items[0][0] in z.keys()
    assert items[0][0] in z
    assert items[0][1] in z.values()
    assert (object(), object()) not in z.items()
    assert object() not in z.keys()
    assert object() not in z
    assert object() not in z.values()


def stress_test_mapping_updates(z: MutableMapping) -> None:
    # Certain mappings shuffle between several underlying stores
    # during updates.  This stress tests the internal mapping
    # consistency.
    r = random.Random(42)

    keys = list(string.ascii_lowercase)
    values = [to_bytestring(s) for s in generate_random_strings(len(keys), 1, 10)]

    z.clear()
    assert len(z) == 0

    for k, v in zip(keys, values):
        z[k] = v
    assert len(z) == len(keys)
    assert sorted(z) == sorted(keys)
    assert sorted(z.items()) == sorted(zip(keys, values))

    for _ in range(3):
        r.shuffle(keys)
        r.shuffle(values)
        for k, v in zip(keys, values):
            z[k] = v
        check_items(z, list(zip(keys, values)))

        r.shuffle(keys)
        r.shuffle(values)
        z.update(zip(keys, values))
        check_items(z, list(zip(keys, values)))


def check_empty_mapping(z: MutableMapping) -> None:
    assert not z
    assert list(z) == list(z.keys()) == []
    assert list(z.values()) == []
    assert list(z.items()) == []
    assert len(z) == 0
    assert "x" not in z
    assert "x" not in z.keys()
    assert ("x", b"123") not in z.items()
    assert b"123" not in z.values()


def check_mapping(z: MutableMapping) -> None:
    """See also test_zip.check_mapping"""
    assert type(z).__name__ in str(z)
    assert type(z).__name__ in repr(z)
    assert isinstance(z, MutableMapping)
    check_empty_mapping(z)

    z["abc"] = b"456"
    z["xyz"] = b"12"
    assert len(z) == 2
    assert z["abc"] == b"456"

    check_items(z, [("abc", b"456"), ("xyz", b"12")])

    assert "abc" in z
    assert "xyz" in z
    assert "def" not in z
    assert object() not in z

    with pytest.raises(KeyError):
        z["def"]

    z.update(xyz=b"707", uvw=b"000")
    check_items(z, [("abc", b"456"), ("xyz", b"707"), ("uvw", b"000")])
    z.update([("xyz", b"654"), ("uvw", b"999")])
    check_items(z, [("abc", b"456"), ("xyz", b"654"), ("uvw", b"999")])
    z.update({"xyz": b"321"})
    check_items(z, [("abc", b"456"), ("xyz", b"321"), ("uvw", b"999")])
    # Update with iterator (can read only once)
    z.update(iter([("foo", b"132"), ("bar", b"887")]))
    check_items(
        z,
        [
            ("abc", b"456"),
            ("xyz", b"321"),
            ("uvw", b"999"),
            ("foo", b"132"),
            ("bar", b"887"),
        ],
    )

    del z["abc"]
    with pytest.raises(KeyError):
        z["abc"]
    with pytest.raises(KeyError):
        del z["abc"]
    assert "abc" not in z
    assert set(z) == {"uvw", "xyz", "foo", "bar"}
    assert len(z) == 4

    z["def"] = b"\x00\xff"
    assert len(z) == 5
    assert z["def"] == b"\x00\xff"
    assert "def" in z

    stress_test_mapping_updates(z)


def check_different_keys_threadsafe(
    z: MutableMapping, allow_keyerror: bool = False
) -> None:
    barrier = threading.Barrier(2)
    counters = [0, 0]

    def worker(idx, key, value):
        barrier.wait()
        while any(c < 10 for c in counters):
            z[key] = value
            try:
                assert z[key] == value
                del z[key]
            except KeyError:
                if allow_keyerror:
                    continue  # Try again, don't inc i
                raise

            assert key not in z
            with pytest.raises(KeyError):
                _ = z[key]
            with pytest.raises(KeyError):
                del z[key]
            assert len(z) in (0, 1)
            counters[idx] += 1

    with ThreadPoolExecutor(2) as ex:
        f1 = ex.submit(worker, 0, "x", b"123")
        f2 = ex.submit(worker, 1, "y", b"456")
        f1.result()
        f2.result()

    assert not z


def check_same_key_threadsafe(z: MutableMapping) -> None:
    barrier = threading.Barrier(4)
    counters = [0, 0, 0, 0]

    def w_set():
        barrier.wait()
        while any(c < 10 for c in counters):
            z["x"] = b"123"
            counters[0] += 1

    def w_update():
        barrier.wait()
        while any(c < 10 for c in counters):
            z.update(x=b"456")
            counters[1] += 1

    def w_del():
        barrier.wait()
        while any(c < 10 for c in counters):
            try:
                del z["x"]
                counters[2] += 1
            except KeyError:
                pass

    def w_get():
        barrier.wait()
        while any(c < 10 for c in counters):
            try:
                assert z["x"] in (b"123", b"456")
                counters[3] += 1
            except KeyError:
                pass

    with ThreadPoolExecutor(4) as ex:
        futures = [
            ex.submit(w_set),
            ex.submit(w_update),
            ex.submit(w_del),
            ex.submit(w_get),
        ]
        for f in futures:
            f.result()

    z.pop("x", None)


def check_closing(z: ZictBase) -> None:
    z.close()


def check_bad_key_types(z: MutableMapping, has_del: bool = True) -> None:
    """z does not accept any Hashable as keys.
    Test that it reacts correctly when confronted with an invalid key type.
    """
    bad = object()

    assert bad not in z
    assert bad not in z.keys()
    assert (bad, b"123") not in z.items()

    with pytest.raises(TypeError):
        z[bad] = b"123"
    with pytest.raises(TypeError):
        z.update({bad: b"123"})
    with pytest.raises(KeyError):
        z[bad]
    if has_del:
        with pytest.raises(KeyError):
            del z[bad]


def check_bad_value_types(z: MutableMapping) -> None:
    """z does not accept any Python object as values.
    Test that it reacts correctly when confronted with an invalid value type.
    """
    bad = object()

    assert bad not in z.values()
    assert ("x", bad) not in z.items()

    with pytest.raises(TypeError):
        z["x"] = bad
    with pytest.raises(TypeError):
        z.update({"x": bad})


class SimpleDict(ZictBase, UserDict):
    def __init__(self):
        ZictBase.__init__(self)
        UserDict.__init__(self)


class SlowDict(UserDict):
    def __init__(self, delay):
        self.delay = delay
        super().__init__(self)

    def __getitem__(self, key):
        time.sleep(self.delay)
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        time.sleep(self.delay)
        super().__setitem__(key, value)



import pickle

import pytest

from zict.common import locked
from zict.tests.utils_test import SimpleDict


def test_close_on_del():
    closed = False

    class D(SimpleDict):
        def close(self):
            nonlocal closed
            closed = True

    d = D()
    del d
    assert closed


def test_context():
    closed = False

    class D(SimpleDict):
        def close(self):
            nonlocal closed
            closed = True

    d = D()
    with d as d2:
        assert d2 is d
    assert closed


def test_update():
    items = []

    class D(SimpleDict):
        def _do_update(self, items_):
            nonlocal items
            items = items_

    d = D()
    d.update({"x": 1})
    assert list(items) == [("x", 1)]
    d.update(iter([("x", 2)]))
    assert list(items) == [("x", 2)]
    d.update({"x": 3}, y=4)
    assert list(items) == [("x", 3), ("y", 4)]
    d.update(x=5)
    assert list(items) == [("x", 5)]

    # Special kwargs can't overwrite positional-only parameters
    d.update(self=1, other=2)
    assert list(items) == [("self", 1), ("other", 2)]


def test_discard():
    class D(SimpleDict):
        def __getitem__(self, key):
            raise AssertionError()

    d = D()
    d["x"] = 1
    d["z"] = 2
    d.discard("x")
    d.discard("y")
    assert d.data == {"z": 2}


def test_pickle():
    d = SimpleDict()
    d["x"] = 1
    d2 = pickle.loads(pickle.dumps(d))
    assert d2.data == {"x": 1}


def test_lock(is_locked):
    class CustomError(Exception):
        pass

    class D(SimpleDict):
        @locked
        def f(self, crash):
            assert is_locked(self)
            with self.unlock():
                assert not is_locked(self)
            assert is_locked(self)

            # context manager re-acquires the lock on failure
            with pytest.raises(CustomError):
                with self.unlock():
                    raise CustomError()
            assert is_locked(self)

            if crash:
                raise CustomError()

    d = D()
    assert not is_locked(d)
    d.f(crash=False)
    assert not is_locked(d)

    # decorator releases the lock on failure
    with pytest.raises(CustomError):
        d.f(crash=True)
    assert not is_locked(d)



from __future__ import annotations

import pathlib
import sys
from collections.abc import Iterable, Iterator
from typing import (  # TODO import from collections.abc (needs Python >=3.9)
    ItemsView,
    ValuesView,
)

from zict.common import ZictBase


def _encode_key(key: str) -> bytes:
    return key.encode("utf-8")


def _decode_key(key: bytes) -> str:
    return key.decode("utf-8")


class LMDB(ZictBase[str, bytes]):
    """Mutable Mapping interface to a LMDB database.

    Keys must be strings, values must be bytes

    Parameters
    ----------
    directory: str
    map_size: int
        On Linux and MacOS, maximum size of the database file on disk.
        Defaults to 1 TiB on 64 bit systems and 1 GiB on 32 bit ones.

        On Windows, preallocated total size of the database file on disk. Defaults to
        10 MiB to encourage explicitly setting it.

    Notes
    -----
    None of this class is thread-safe - not even normally trivial methods such as
    ``__len__ `` or ``__contains__``.

    Examples
    --------
    >>> z = LMDB('/tmp/somedir/')  # doctest: +SKIP
    >>> z['x'] = b'123'  # doctest: +SKIP
    >>> z['x']  # doctest: +SKIP
    b'123'
    """

    def __init__(self, directory: str | pathlib.Path, map_size: int | None = None):
        import lmdb

        super().__init__()
        if map_size is None:
            if sys.platform != "win32":
                map_size = min(2**40, sys.maxsize // 4)
            else:
                map_size = 10 * 2**20

        self.db = lmdb.open(
            str(directory),
            subdir=True,
            map_size=map_size,
            sync=False,
            writemap=True,
        )

    def __getitem__(self, key: str) -> bytes:
        if not isinstance(key, str):
            raise KeyError(key)
        with self.db.begin() as txn:
            value = txn.get(_encode_key(key))
        if value is None:
            raise KeyError(key)
        return value

    def __setitem__(self, key: str, value: bytes) -> None:
        if not isinstance(key, str):
            raise TypeError(key)
        if not isinstance(value, bytes):
            raise TypeError(value)
        with self.db.begin(write=True) as txn:
            txn.put(_encode_key(key), value)

    def __contains__(self, key: object) -> bool:
        if not isinstance(key, str):
            return False
        with self.db.begin() as txn:
            return txn.cursor().set_key(_encode_key(key))

    def __iter__(self) -> Iterator[str]:
        cursor = self.db.begin().cursor()
        return (_decode_key(k) for k in cursor.iternext(keys=True, values=False))

    def items(self) -> ItemsView[str, bytes]:
        return LMDBItemsView(self)

    def values(self) -> ValuesView[bytes]:
        return LMDBValuesView(self)

    def _do_update(self, items: Iterable[tuple[str, bytes]]) -> None:
        # Optimized version of update() using a single putmulti() call.
        items_enc = []
        for key, value in items:
            if not isinstance(key, str):
                raise TypeError(key)
            if not isinstance(value, bytes):
                raise TypeError(value)
            items_enc.append((_encode_key(key), value))

        with self.db.begin(write=True) as txn:
            consumed, added = txn.cursor().putmulti(items_enc)
            assert consumed == added == len(items_enc)

    def __delitem__(self, key: str) -> None:
        if not isinstance(key, str):
            raise KeyError(key)
        with self.db.begin(write=True) as txn:
            if not txn.delete(_encode_key(key)):
                raise KeyError(key)

    def __len__(self) -> int:
        return self.db.stat()["entries"]

    def close(self) -> None:
        self.db.close()


class LMDBItemsView(ItemsView[str, bytes]):
    _mapping: LMDB  # FIXME CPython implementation detail
    __slots__ = ()

    def __contains__(self, item: object) -> bool:
        key: str
        value: object
        key, value = item  # type: ignore
        try:
            v = self._mapping[key]
        except KeyError:
            return False
        else:
            return v == value

    def __iter__(self) -> Iterator[tuple[str, bytes]]:
        cursor = self._mapping.db.begin().cursor()
        return ((_decode_key(k), v) for k, v in cursor.iternext(keys=True, values=True))


class LMDBValuesView(ValuesView[bytes]):
    _mapping: LMDB  # FIXME CPython implementation detail
    __slots__ = ()

    def __contains__(self, value: object) -> bool:
        return any(value == v for v in self)

    def __iter__(self) -> Iterator[bytes]:
        cursor = self._mapping.db.begin().cursor()
        return cursor.iternext(keys=False, values=True)



import os
import pathlib
import sys

import pytest

from zict import File
from zict.tests import utils_test


def test_mapping(tmp_path, check_fd_leaks):
    """
    Test mapping interface for File().
    """
    z = File(tmp_path)
    utils_test.check_mapping(z)


@pytest.mark.parametrize("dirtype", [str, pathlib.Path, lambda x: x])
def test_implementation(tmp_path, check_fd_leaks, dirtype):
    z = File(dirtype(tmp_path))
    assert not z

    z["x"] = b"123"
    assert os.listdir(tmp_path) == ["x#0"]
    with open(tmp_path / "x#0", "rb") as f:
        assert f.read() == b"123"

    assert "x" in z
    out = z["x"]
    assert isinstance(out, bytearray)
    assert out == b"123"


def test_memmap_implementation(tmp_path, check_fd_leaks):
    z = File(tmp_path, memmap=True)
    assert not z

    mv = memoryview(b"123")
    assert "x" not in z
    z["x"] = mv
    assert os.listdir(tmp_path) == ["x#0"]
    assert "x" in z
    mv2 = z["x"]
    assert mv2 == b"123"
    # Buffer is writeable
    mv2[0] = mv2[1]
    assert mv2 == b"223"


def test_str(tmp_path, check_fd_leaks):
    z = File(tmp_path)
    assert str(z) == repr(z) == f"<File: {tmp_path}, 0 elements>"


def test_setitem_typeerror(tmp_path, check_fd_leaks):
    z = File(tmp_path)
    with pytest.raises(TypeError):
        z["x"] = 123


def test_contextmanager(tmp_path, check_fd_leaks):
    with File(tmp_path) as z:
        z["x"] = b"123"

    with open(tmp_path / "x#0", "rb") as fh:
        assert fh.read() == b"123"


def test_delitem(tmp_path, check_fd_leaks):
    z = File(tmp_path)

    z["x"] = b"123"
    assert os.listdir(tmp_path) == ["x#0"]
    del z["x"]
    assert os.listdir(tmp_path) == []
    # File name is never repeated
    z["x"] = b"123"
    assert os.listdir(tmp_path) == ["x#1"]
    # __setitem__ deletes the previous file
    z["x"] = b"123"
    assert os.listdir(tmp_path) == ["x#2"]


def test_missing_key(tmp_path, check_fd_leaks):
    z = File(tmp_path)

    with pytest.raises(KeyError):
        z["x"]


def test_arbitrary_chars(tmp_path, check_fd_leaks):
    z = File(tmp_path)

    # Avoid hitting the Windows max filename length
    chunk = 16
    for i in range(1, 128, chunk):
        key = "".join(["foo_"] + [chr(i) for i in range(i, min(128, i + chunk))])
        with pytest.raises(KeyError):
            z[key]
        z[key] = b"foo"
        assert z[key] == b"foo"
        assert list(z) == [key]
        assert list(z.keys()) == [key]
        assert list(z.items()) == [(key, b"foo")]
        assert list(z.values()) == [b"foo"]

        zz = File(tmp_path)
        assert zz[key] == b"foo"
        assert list(zz) == [key]
        assert list(zz.keys()) == [key]
        assert list(zz.items()) == [(key, b"foo")]
        assert list(zz.values()) == [b"foo"]
        del zz

        del z[key]
        with pytest.raises(KeyError):
            z[key]


def test_write_list_of_bytes(tmp_path, check_fd_leaks):
    z = File(tmp_path)

    z["x"] = [b"123", b"4567"]
    assert z["x"] == b"1234567"


def test_bad_types(tmp_path, check_fd_leaks):
    z = File(tmp_path)
    utils_test.check_bad_key_types(z)
    utils_test.check_bad_value_types(z)


@pytest.mark.stress
@pytest.mark.repeat(utils_test.REPEAT_STRESS_TESTS)
def test_stress_different_keys_threadsafe(tmp_path):
    z = File(tmp_path)
    utils_test.check_different_keys_threadsafe(z)
    utils_test.check_mapping(z)


@pytest.mark.stress
@pytest.mark.repeat(utils_test.REPEAT_STRESS_TESTS)
@pytest.mark.skipif(sys.platform == "win32", reason="Can't delete file with open fd")
def test_stress_same_key_threadsafe(tmp_path):
    z = File(tmp_path)
    utils_test.check_same_key_threadsafe(z)
    utils_test.check_mapping(z)



import gc
import threading
from collections import UserDict
from concurrent.futures import ThreadPoolExecutor

import pytest

from zict import Cache, WeakValueMapping
from zict.tests import utils_test


def test_cache_get_set_del():
    d = Cache({}, {})

    # getitem (bad key)
    with pytest.raises(KeyError):
        d[0]
    assert (d.data, d.cache) == ({}, {})

    # setitem(no update); cache is empty
    d[1] = 10
    assert (d.data, d.cache) == ({1: 10}, {1: 10})

    # getitem; cache is full
    assert d[1] == 10
    assert (d.data, d.cache) == ({1: 10}, {1: 10})

    # getitem; cache is empty
    d.cache.clear()
    assert d[1] == 10
    assert (d.data, d.cache) == ({1: 10}, {1: 10})

    # setitem(update); cache is full
    d[1] = 20
    assert (d.data, d.cache) == ({1: 20}, {1: 20})

    # setitem(update); cache is empty
    d.cache.clear()
    d[1] = 30
    assert (d.data, d.cache) == ({1: 30}, {1: 30})

    # delitem; cache is full
    del d[1]
    assert (d.data, d.cache) == ({}, {})

    # delitem; cache is empty
    d[1] = 10
    d.cache.clear()
    del d[1]
    assert (d.data, d.cache) == ({}, {})

    # delitem (bad key)
    with pytest.raises(KeyError):
        del d[0]
    assert (d.data, d.cache) == ({}, {})


def test_do_not_read_from_data():
    """__len__, __iter__, __contains__, and keys() do not populate the cache"""

    class D(UserDict):
        def __getitem__(self, key):
            raise AssertionError()

    d = Cache(D({1: 10, 2: 20}), {})
    assert len(d) == 2
    assert list(d) == [1, 2]
    assert 1 in d
    assert 3 not in d
    assert d.keys() == {1, 2}
    assert d.cache == {}


def test_no_update_on_set():
    d = Cache({}, {}, update_on_set=False)
    d[1] = 10
    assert (d.data, d.cache) == ({1: 10}, {})
    assert d[1] == 10
    assert (d.data, d.cache) == ({1: 10}, {1: 10})
    d[1] = 20
    assert (d.data, d.cache) == ({1: 20}, {})
    assert d[1] == 20
    assert (d.data, d.cache) == ({1: 20}, {1: 20})


def test_slow_fails():
    """data.__setitem__ raises; e.g. disk full"""

    class D(UserDict):
        def __setitem__(self, key, value):
            if value == "fail":
                self.pop(key, None)
                raise ValueError()
            super().__setitem__(key, value)

    d = Cache(D(), {})

    # setitem(no update); cache is empty
    with pytest.raises(ValueError):
        d[1] = "fail"
    assert (d.data.data, d.cache) == ({}, {})

    # setitem(update); cache is empty
    d[1] = 10
    d.cache.clear()
    assert (d.data.data, d.cache) == ({1: 10}, {})
    with pytest.raises(ValueError):
        d[1] = "fail"
    assert (d.data.data, d.cache) == ({}, {})

    # setitem(update); cache is full
    d[1] = 10
    assert (d.data.data, d.cache) == ({1: 10}, {1: 10})
    with pytest.raises(ValueError):
        d[1] = "fail"
    assert (d.data.data, d.cache) == ({}, {})


def test_weakvaluemapping():
    class C:
        pass

    d = WeakValueMapping()
    a = C()
    d["a"] = a
    assert d["a"] is a
    del a
    gc.collect()  # Needed in pypy
    assert "a" not in d

    # str does not support weakrefs
    b = "bbb"
    d["b"] = b
    assert "b" not in d


def test_mapping():
    """
    Test mapping interface for Cache().
    """
    buff = Cache({}, {})
    utils_test.check_mapping(buff)
    utils_test.check_closing(buff)

    buff.clear()
    assert not buff.cache
    assert not buff.data
    assert not buff._last_updated


@pytest.mark.parametrize("get_when", ("before", "after"))
@pytest.mark.parametrize("set_when", ("before", "after"))
@pytest.mark.parametrize(
    "starts_first,seed,update_on_set",
    [
        ("get", True, False),
        ("set", False, False),
        ("set", False, True),
        ("set", True, False),
        ("set", True, True),
    ],
)
@pytest.mark.parametrize("ends_first", ("get", "set"))
def test_multithread_race_condition_set_get(
    get_when, set_when, starts_first, seed, update_on_set, ends_first
):
    """Test race conditions between __setitem__ and __getitem__ on the same key"""
    in_get = threading.Event()
    block_get = threading.Event()
    in_set = threading.Event()
    block_set = threading.Event()

    class Slow(UserDict):
        def __getitem__(self, k):
            if get_when == "before":
                try:
                    v = self.data[k]
                finally:
                    in_get.set()
                assert block_get.wait(timeout=5)
                return v
            else:
                in_get.set()
                assert block_get.wait(timeout=5)
                return self.data[k]

        def __setitem__(self, k, v):
            if set_when == "before":
                self.data[k] = v
                in_set.set()
                assert block_set.wait(timeout=5)
            else:
                in_set.set()
                assert block_set.wait(timeout=5)
                self.data[k] = v

    z = Cache(Slow(), {}, update_on_set=update_on_set)
    if seed:
        block_set.set()
        z["x"] = 1
        in_set.clear()
        block_set.clear()
        assert z.data.data == {"x": 1}
        assert set(z._last_updated) == {"x"}
        if update_on_set:
            assert z.cache == {"x": 1}
        else:
            assert z.cache == {}

    with ThreadPoolExecutor(2) as ex:
        if starts_first == "get":
            get_fut = ex.submit(z.__getitem__, "x")
            assert in_get.wait(timeout=5)
            set_fut = ex.submit(z.__setitem__, "x", 2)
            assert in_set.wait(timeout=5)
        else:
            set_fut = ex.submit(z.__setitem__, "x", 2)
            assert in_set.wait(timeout=5)
            get_fut = ex.submit(z.__getitem__, "x")
            assert in_get.wait(timeout=5)

        if ends_first == "get":
            block_get.set()
            try:
                assert get_fut.result() in (1, 2)
            except KeyError:
                pass
            block_set.set()
            set_fut.result()
        else:
            block_set.set()
            set_fut.result()
            block_get.set()
            try:
                assert get_fut.result() in (1, 2)
            except KeyError:
                pass

    assert z.data.data == {"x": 2}
    # The cache is either not populated or up to date
    assert z.cache in ({}, {"x": 2})
    assert set(z._last_updated) == {"x"}


@pytest.mark.parametrize("get_when", ("before", "after"))
def test_multithread_race_condition_del_get(get_when):
    """Test race conditions between __delitem__ and __getitem__ on the same key"""
    in_get = threading.Event()
    block_get = threading.Event()

    class Slow(UserDict):
        def __getitem__(self, k):
            if get_when == "before":
                v = self.data[k]
                in_get.set()
                assert block_get.wait(timeout=5)
                return v
            else:
                in_get.set()
                assert block_get.wait(timeout=5)
                return self.data[k]

    z = Cache(Slow(), {}, update_on_set=False)
    z["x"] = 1
    assert z.data.data == {"x": 1}
    assert set(z._last_updated) == {"x"}
    assert z.cache == {}

    with ThreadPoolExecutor(1) as ex:
        get_fut = ex.submit(z.__getitem__, "x")
        assert in_get.wait(timeout=5)
        del z["x"]

        block_get.set()
        if get_when == "before":
            assert get_fut.result() == 1
        else:
            with pytest.raises(KeyError):
                get_fut.result()

    assert not z.data.data
    assert not z.cache
    assert not z._last_updated


@pytest.mark.parametrize("set_when", ("before", "after"))
@pytest.mark.parametrize(
    "seed_data,seed_cache", [(False, False), (True, False), (True, True)]
)
@pytest.mark.parametrize("update_on_set", [False, True])
def test_multithread_race_condition_del_set(
    set_when, seed_data, seed_cache, update_on_set
):
    """Test race conditions between __delitem__ and __setitem__ on the same key"""
    in_set = threading.Event()
    block_set = threading.Event()

    class Slow(UserDict):
        def __setitem__(self, k, v):
            if set_when == "before":
                self.data[k] = v
                in_set.set()
                assert block_set.wait(timeout=5)
            else:
                in_set.set()
                assert block_set.wait(timeout=5)
                self.data[k] = v

    z = Cache(Slow(), {}, update_on_set=update_on_set)
    if seed_data:
        block_set.set()
        z["x"] = 1
        in_set.clear()
        block_set.clear()
        if seed_cache and not update_on_set:
            _ = z["x"]

    with ThreadPoolExecutor(1) as ex:
        set_fut = ex.submit(z.__setitem__, "x", 2)
        assert in_set.wait(timeout=5)
        try:
            del z["x"]
        except KeyError:
            pass
        block_set.set()
        set_fut.result()

    assert z.data.data in ({"x": 2}, {})
    assert z.cache in (z.data.data, {})
    assert z._last_updated.keys() == z.data.data.keys()


@pytest.mark.parametrize("set1_when", ("before", "after"))
@pytest.mark.parametrize("set2_when", ("before", "after"))
@pytest.mark.parametrize("starts_first", (1, 2))
@pytest.mark.parametrize("ends_first", (1, 2))
@pytest.mark.parametrize(
    "seed_data,seed_cache", [(False, False), (True, False), (True, True)]
)
@pytest.mark.parametrize("update_on_set", [False, True])
def test_multithread_race_condition_set_set(
    set1_when, set2_when, starts_first, ends_first, seed_data, seed_cache, update_on_set
):
    """Test __setitem__ in race condition with itself"""
    when = {1: set1_when, 2: set2_when}
    in_set = {1: threading.Event(), 2: threading.Event()}
    block_set = {1: threading.Event(), 2: threading.Event()}

    class Slow(UserDict):
        def __setitem__(self, k, v):
            if v == 0:
                # seed
                self.data[k] = v
                return

            if when[v] == "before":
                self.data[k] = v
                in_set[v].set()
                assert block_set[v].wait(timeout=5)
            else:
                in_set[v].set()
                assert block_set[v].wait(timeout=5)
                self.data[k] = v

    z = Cache(Slow(), {}, update_on_set=update_on_set)
    if seed_data:
        z["x"] = 0
        if seed_cache and not update_on_set:
            _ = z["x"]

    with ThreadPoolExecutor(2) as ex:
        futures = {}
        futures[starts_first] = ex.submit(z.__setitem__, "x", starts_first)
        assert in_set[starts_first].wait(timeout=5)
        starts_second = 2 if starts_first == 1 else 1
        futures[starts_second] = ex.submit(z.__setitem__, "x", starts_second)
        assert in_set[starts_second].wait(timeout=5)

        block_set[ends_first].set()
        futures[ends_first].result()
        ends_second = 2 if ends_first == 1 else 1
        block_set[ends_second].set()
        futures[ends_second].result()

    assert z.data.data in ({"x": 1}, {"x": 2})
    assert z.cache in (z.data.data, {})
    assert set(z._last_updated) == {"x"}


@pytest.mark.stress
@pytest.mark.repeat(utils_test.REPEAT_STRESS_TESTS)
@pytest.mark.parametrize("update_on_set", [False, True])
def test_stress_different_keys_threadsafe(update_on_set):
    buff = Cache(utils_test.SlowDict(0.001), {}, update_on_set=update_on_set)
    utils_test.check_different_keys_threadsafe(buff)
    assert not buff.cache
    assert not buff.data
    assert not buff._last_updated
    utils_test.check_mapping(buff)


@pytest.mark.stress
@pytest.mark.repeat(utils_test.REPEAT_STRESS_TESTS)
@pytest.mark.parametrize("update_on_set", [False, True])
def test_stress_same_key_threadsafe(update_on_set):
    buff = Cache(utils_test.SlowDict(0.001), {}, update_on_set=update_on_set)
    utils_test.check_same_key_threadsafe(buff)
    assert not buff.cache
    assert not buff.data
    assert not buff._last_updated
    utils_test.check_mapping(buff)
